module OrderedString = struct
   type t = string
   let compare = compare
end
module StringMap = Map.Make(OrderedString)

type expr =
  | OID of string
  | IDENT of string
  | INT of int
  | FLOAT of float
  | FACT of string * expr list
  | ASSOC of string * string * string

type oid_tbl = expr StringMap.t
type fact_tbl = expr StringMap.t
type tbl_map = fact_tbl StringMap.t
type facts_data = oid_tbl * tbl_map * (string StringMap.t)

let rec concat = function
  | [] -> ""
  | s::[] -> s
  | s::ss -> s^" "^(concat ss)

let rec string_of_expr (expr:expr) : string =
  match expr with
    | OID(oid) -> oid
    | IDENT(id) -> id
    | INT(i) -> string_of_int i
    | FLOAT(f) -> Printf.sprintf "%.2f" f
    | ASSOC(assoc,factName,keyName) ->
	"(assoc "^assoc^" "^factName^" "^keyName^")"
    | FACT(id,ee) -> id^" "^(concat (List.map string_of_expr ee))

let print (facts:expr list) : unit =
  let p_fact f = output_string stdout ((string_of_expr f)^"\n")
  in List.iter p_fact facts

let name_of_expr = function
  | FACT(name,_) -> name
  | IDENT(name) -> name
  | OID(oid) -> oid
  | INT(i) -> string_of_int i
  | FLOAT(f) -> Printf.sprintf "%.2f" f
  | ASSOC(name,key,value) -> name

(* lookup a fact in the OID map *)
let get_fact (facts:facts_data) (oid:string) : expr =
  let oid_map,_,_ = facts
  in StringMap.find oid oid_map

let primaryKeyName (facts:facts_data) factname : string =
  let _,_,prim_key_map = facts
  in StringMap.find factname prim_key_map

let primaryKey (facts:facts_data) fact : string =
  let _,_,prim_key_map = facts in
  (* raises Not_found if there is no primary key associated *)
  let factname = name_of_expr fact in
  let prim_key_name = primaryKeyName facts factname in
  let rec f (p:string) (expr:expr) : string =
    match expr with
      | FACT(name,expr) ->
	  if name=prim_key_name then name_of_expr (List.hd expr)
	  else List.fold_left f p expr
      | _ -> p
  in f "" fact

let field_value (facts:facts_data) (key:string) (oid:string) : string =
  try 
    let fact = get_fact facts oid in
    let rec f (p:string) (expr:expr) : string =
      match expr with
	| FACT(name,expr) ->
	    if name=key then name_of_expr (List.hd expr)
	    else List.fold_left f p expr
	| _ -> p
    in f oid fact
  with Not_found -> "##"^oid^"##"

let organize_facts facts =
  let prim_key_map =
    let fold_fact map fact = match fact with
      | ASSOC(keytype,key,value) ->
	  if keytype="primaryKeyAssociation"
	  then StringMap.add key value map
	  else map
      | _ -> map
    in List.fold_left fold_fact StringMap.empty facts in
  let (tbl_map:tbl_map),(oid_map:oid_tbl) =
    let f (tbl_map,oid_map) fact = match fact with
      | FACT(factname,_) -> 
	  (try
	     let key:string = primaryKey (oid_map,tbl_map,prim_key_map) fact
	     in
	       (try
		  let facttbl =
		    let oldmap = StringMap.find factname tbl_map
		    in StringMap.add key fact oldmap
		  in
		    (StringMap.add factname facttbl tbl_map),
		    (StringMap.add key fact oid_map)
		with Not_found ->
		  let facttbl = StringMap.add key fact StringMap.empty
		  in
		    (StringMap.add factname facttbl tbl_map),
		    (StringMap.add key fact oid_map))
           (* if no primary key, leave fact out of database *)
	   with Not_found -> tbl_map,oid_map)
      | _ -> tbl_map,oid_map
    in List.fold_left f (StringMap.empty,StringMap.empty) facts
  in oid_map,tbl_map,prim_key_map

let tbl_names (facts:facts_data) : string list =
  let _,tbl_map,_ = facts in
  let f key data accum = key::accum
  in StringMap.fold f tbl_map []

let get_rows (tbl_name:string) (tbl_map:tbl_map) : expr list =
  let tbl = StringMap.find tbl_name tbl_map in
  let f key data ll = data::ll
  in StringMap.fold f tbl []

