open Vtl
open Vtl_server
open Facts

let parse_facts (chan:in_channel) : expr list =
  let parse scan_f parse_f pos_f =
    let position (line,col) = ("near line "
                               ^ (string_of_int line)
                               ^ ", column "
                               ^ (string_of_int col)) in
    let lexbuf = Lexing.from_channel chan
    in
      try
	parse_f scan_f lexbuf
      with exn ->
	let pos = position (pos_f lexbuf) in
          begin match exn with
	    | Failure(msg) ->
		failwith ("program failure, while parsing " ^ pos ^ ": " ^ msg)
            | Parsing.Parse_error -> failwith("syntax error " ^ pos)
            | Sys_error msg -> failwith ("I/O error: " ^ msg)
            | _ -> raise exn
	  end
  in
    parse Facts_lexer.token Facts_parser.root Facts_lexer.position

(* home page. VTL: none *)
let vtl_home (facts:facts_data) : service =
  let svc (path,(params:params)) = empty
  in ("home.html",svc)

let groups (facts:facts_data) : Vtl.value =
  let _,tbl_map,_ = facts in
  let groupnames = tbl_names facts in
  let f groups group =
    let factdetails = get_rows group tbl_map
    in Vtl.ASSOC(extend (extend empty (REF "name") (STRING group))
		   (REF "length") (Vtl.INT (List.length factdetails)))::groups
  in  LIST(List.fold_left f [] groupnames)
  
(* factgroup page. VRL: $groupnames = LIST of group name strings *)
let vtl_factgroup (facts:facts_data) : service =
  let svc (path,(params:params)) =
    Vtl.extend empty (REF "$groups") (groups facts)
  in ("factgroup.html",svc)

let str_val (x:Facts.expr) : Vtl.value = Vtl.STRING(Facts.string_of_expr x)

let func_val (x:expr) : Vtl.value = match x with
  | FACT(ident,expr) -> Vtl.STRING(ident)
  | x -> str_val x

let param_val (facts:facts_data) (key:string) (x:expr) : Vtl.value =
  let env = empty
  in 
    match x with
      | FACT(ident,[OID(oid)]) ->
	  let env = extend env (REF "data")
		      (if key=ident then Vtl.STRING oid
		       else (Vtl.STRING (field_value facts "name" oid))) in
	  let env =
	    if key=ident then (* this is the primary key *)
	      extend env (REF "name") (Vtl.STRING oid)
	    else
	      try
		let group = func_val (get_fact facts oid) in
		let env = extend env (DOTREF("link",REF("group"))) group
		in extend env (DOTREF("link",REF("anchor"))) (Vtl.STRING oid)
	      with Not_found -> env
	  in Vtl.ASSOC(env)
      | FACT(ident,expr) ->
	  Vtl.ASSOC(extend env (REF "data") (str_val (FACT("",expr))))
      | x -> Vtl.ASSOC(extend env (REF"data") (str_val x))

let param_as_string expr : string =
  match expr with
    | FACT(_,x) -> string_of_expr (List.hd x)
    | x -> string_of_expr x

let fact_name expr =
  match expr with
    | FACT(ident,_) -> ident
    | x -> ""

let fact_expr expr =
  match expr with
    | FACT(_,expr) -> expr
    | x -> []

(* fact page. VRL: $group.name: string, $group.facts: fact list *)
let vtl_fact (facts:facts_data) : service =
  let _,tbl_map,_ = facts in
  let svc (path,(params:params)) =
    let fmt_fact fact : Vtl.value =
      match fact with
	| FACT(name,exprlist) ->
	    let key = primaryKeyName facts name
	    in LIST(List.map (param_val facts key) exprlist)
	| _ -> LIST[] in
    let group = Vtl_server.getEnv params "group" in
    let factdetail =
      let sortname = Vtl_server.getEnv params "sort" in
      let rows = get_rows group tbl_map in
      let keyname = primaryKeyName facts group in
      let f r1 r2 =
	let c1 = field_value facts sortname (primaryKey facts r1) in
	let c2 = field_value facts sortname (primaryKey facts r2)
	in compare c1 c2
      in List.sort f rows in
    let headersvalue = match List.hd factdetail with
      | FACT(name,expr) -> LIST(List.map func_val expr)
      | _ -> LIST([]) in
    (*
      * set the ENV properties
    *)
    let factsvalue = LIST(List.map fmt_fact factdetail) in
    let env = Vtl.extend empty (DOTREF("$group",REF("name"))) (STRING group) in
    let env = Vtl.extend env (DOTREF("$group",REF("facts"))) factsvalue in
    let env = Vtl.extend env (DOTREF("$group",REF("headers"))) headersvalue
    in Vtl.extend env (REF "$groups") (groups facts)
  in ("fact.html",svc)

let services (facts:facts_data) : services =
  [("",(vtl_home facts));
   ("factgroup",(vtl_factgroup facts));
   ("fact",(vtl_fact facts))]

let not_found_svc : service =
  let svc(path,(params:params)) =
    let uri =
      path^"?"^(List.fold_left (fun s (n,v) -> s^n^"="^v^"&") "" params) in
    let env = extend empty (REF "$uri") (STRING uri)
    in env
  in "file_not_found.html",svc

let main () =
  let root = ref "." in
  let port = ref "8042" in
  let f arg value = arg := value in
  let speclist =
    [("-root", Arg.String(f root),"TEMPLATE_ROOT, default="^(!root));
     ("-port", Arg.String(f port),"HTTP port, default="^(!port))]
  in
    try
      Arg.parse speclist (fun x -> ()) "Usage: ";
      let facts = parse_facts stdin in
      let facts_data = organize_facts facts
      in Vtl_server.server (int_of_string !port)
	   !root (services facts_data) not_found_svc
    with Failure(msg) -> Printf.eprintf "%s\n" msg

let _ = Printexc.catch main ()
