open Validate
open Vtl
open Vtl_engine
open Printf

let parse chan scan_f parse_f pos_f =
  let position (line,col) = ("near line "
                             ^ (string_of_int line)
                             ^ ", column "
                             ^ (string_of_int col)) in
  let lexbuf = Lexing.from_channel chan
  in
    try
      parse_f scan_f lexbuf
    with exn ->
      let pos = position (pos_f lexbuf) in
        begin match exn with
	  | Failure(msg) ->
	      failwith ("program error, while parsing " ^ pos ^ ": " ^ msg)
          | Parsing.Parse_error -> failwith("syntax error " ^ pos)
          | Sys_error msg -> failwith ("I/O error: " ^ msg)
          | _ -> raise exn
	end

let rec mergeStatus status = function
    [] -> status
  | f::fs -> let result = f() in
      match result with
	  PASSED -> mergeStatus PASSED fs
	| _ -> result

(* test the environment *)
let va = REF("$a")
let vbc = DOTREF("$b",(REF "c"))
let vbd = DOTREF("$b",(REF "d"))
let vbe = DOTREF("$b",(REF "e"))
let vz = REF("$z")
let vzx = DOTREF("$z",REF("x"))
let vzy = DOTREF("$z",REF("y"))
let vxyz = DOTREF("$x",DOTREF("y",(REF "z")))
let vzwx = DOTREF("$z",DOTREF("w",(REF "x")))

let test_env =
  let env = extendlist empty [va,   (STRING "a");
			      vbc,  (STRING "bc");
			      vzx,  (STRING "zx");
			      vzy,  (STRING "zy");
			      vxyz, (INT 3)] in
  let env = extend env vbd (LIST[INT(1);INT(2);INT(3)]) in (* $b.d *)
  let env = extend env vbe (LIST[STRING("Haley");STRING("Geordie")])
  in env

(* test environment values via string_of_var *)
let f_env () =
  let checkTrivial() =
    Validate.expectString("a",(string_of_var test_env va),"VTL:f_env") in
  let checkNested() =
    Validate.expectString("bc",(string_of_var test_env vbc),"VTL:f_env") in
  let checkExtendNested1() =
    Validate.expectString("zx",(string_of_var test_env vzx),"VTL:f_env") in
  let checkExtendNested2() =
    Validate.expectString("zy",(string_of_var test_env vzy),"VTL:f_env") in
  let checkVeryNested() =
    Validate.expectString("3",(string_of_var test_env vxyz),"VTL:f_env") in
  let checkRebind() =
    (* rebind z.x *)
    let env = extend test_env vzx (STRING "zx2") in
      Validate.expectString("zx2",(string_of_var env vzx),"VTL:f_env") in
  let checkList() =
    let s = string_of_var test_env vbd
    in Validate.expectString("$b.d:list",s,"VTL:f_env") in
  let checkAssoc() =
    let s = string_of_var test_env vz
    in Validate.expectString("$z:assoc",s,"VTL:f_env")

  in mergeStatus PASSED [checkTrivial; checkNested;
			 checkExtendNested1; checkExtendNested2;
			 checkVeryNested; checkRebind; checkList;
			 checkAssoc
			]

(* test name_of_var function *)
let f_name_of_var () =
  let checkTrivial() =
    Validate.expectString("$x",(name_of_var (REF "$x")),"Vtl.name_of_var") in
  let checkNested() =
    let name = name_of_var (DOTREF("$x",DOTREF("y",REF("z")))) in
      Validate.expectString("$x.y.z",name,"Vtl.name_of_var")
  in mergeStatus PASSED [checkTrivial; checkNested]

let test_template = [
  SCHMOO("This is the first.\n");
  BLOCK[SCHMOO("#begin\n");SCHMOO(" b1\n");SCHMOO(" b2\n");SCHMOO("#end\n")];
  SET_STMT(REF("$xyzzy"),ADD(PRIM(INT_EXPR(2)),PRIM(INT_EXPR(3))));
  IF_STMT(EQ(PRIM(REF_EXPR(REF("$xyzzy"))),PRIM(INT_EXPR(42))),
	  SCHMOO("this is wrong. 'xyzzy' is 5.\n"),
	  Some(SCHMOO("this is right.\n")));
  SET_STMT(REF("$xyzzy"),PRIM(INT_EXPR(42)));
  IF_STMT(EQ(PRIM(REF_EXPR(REF("$xyzzy"))),PRIM(INT_EXPR(42))),
	  SCHMOO("this is right.\n"),
	  Some(SCHMOO("this is wrong. 'xyzzy' is 42.\n")));
  SCHMOO("This is the middle.\n");
  SCHMOO("$xyzzy = >>");
  REF_STMT(REF("$xyzzy"));
  SCHMOO("<<\n");
  (* test that loop counters are local only *)
  SET_STMT(DOTREF("$i",REF("foo")),PRIM(INT_EXPR(10)));
  FOR_STMT("$i",REF_EXPR(vbd),
	   BLOCK[SCHMOO("The ball bounced ");
		 REF_STMT(REF("$i"));
		 SCHMOO(" times over ");
		 FOR_STMT("$j",REF_EXPR(vbe),
			  BLOCK[REF_STMT(REF("$j"));
				SCHMOO(" and ")]);
		 SCHMOO(".\n")
		]);
  SCHMOO("This is the last.\n")
]

let expected_document =
"This is the first.
#begin
 b1
 b2
#end
this is right.
this is right.
This is the middle.
$xyzzy = >>42<<
The ball bounced 1 times over Haley and Geordie and .
The ball bounced 2 times over Haley and Geordie and .
The ball bounced 3 times over Haley and Geordie and .
This is the last.
"

let f_eval_template () =
  let doc = eval_template test_env test_template
  in Validate.expectString(expected_document,doc,"Vtl.eval_template")

let f_eval_parser () =
  try
    let template = Vtl_engine.parse_template (open_in "test.vtl")
    in List.iter Vtl.print_stmt template; PASSED
  with Failure(msg) -> FAILED(Some(msg))

(* run the engine. This is the big banana *)
let f_eval_run () =
  try
    let doc = Vtl_engine.run test_env "." "test.vtl"
    in output_string stdout doc; PASSED
  with Failure(msg) -> FAILED(Some(msg))

(* get multiple copies of an environment parameter *)
let f_getEnvAll () =
  let env = [("x","x1"); ("y","y1"); ("x","x2"); ("x","x3"); ("x","x4")] in
  let expect = ["x1";"x2";"x3";"x4"] in
  let actual = Vtl_server.getEnvAll env "x"
  in
    if expect=actual then PASSED
    else FAILED(None)

(* exported values... *)
let alltests = [TEST(f_name_of_var,"vtl.ml:name_of_var");
		TEST(f_env,"vtl.ml:environment");
		TEST(f_eval_template, "vtl_engine:eval_template");
		TEST(f_eval_parser, "vtl_engine:parse_template");
		TEST(f_eval_run, "vtl_engine:run");
		TEST(f_getEnvAll, "vtl_server:getEnvAll")
	       ]

