type id = string
type reference =
  | REF of id
  | DOTREF of id * reference

type value =
  | LIST of value list
  | STRING of string
  | BOOL of bool
  | INT of int
  | ASSOC of env
and env = (string * value) list

type text = string

type range_expr =
  | REF_RANGE of reference
  | INT_RANGE of int

type prim_expr =
  | REF_EXPR of reference
  | STR_EXPR of string
  | INT_EXPR of int
  | BOOL_EXPR of bool
  | LIST_EXPR of prim_expr list
  | RANGE_EXPR of range_expr * range_expr

type expr =
  | PRIM of prim_expr
  | EQ of expr * expr
  | NE of expr * expr
  | GT of expr * expr
  | GE of expr * expr
  | LT of expr * expr
  | LE of expr * expr
  | AND of expr * expr
  | OR of expr * expr
  | ADD of expr * expr
  | SUB of expr * expr
  | MUL of expr * expr
  | DIV of expr * expr
  | MOD of expr * expr
  | NOT of expr

type stmt =
  | SCHMOO of text
  | BLOCK of stmt list
  | IF_STMT of expr * stmt * stmt option
  | FOR_STMT of id * prim_expr * stmt
  | SET_STMT of reference * expr
  | REF_STMT of reference
  | INC_STMT of prim_expr list (* list and int types are bogus here *)
  | PARSE_STMT of prim_expr (* list and string types are bogus here *)

type template = stmt list

(*
 * The Environment is implemented with an associative list,
 * for which the standard library has a nice implementation.
 * See http://caml.inria.fr/ocaml/htmlman/manual048.html,
 * Associative Lists under Module List.
 *)
(* the initial environment has no bindings *)
let empty : env = []

let print s = output_string stdout s

let rec name_of_var (var:reference) =
  match var with
    | REF(id) -> id
    | DOTREF(id,more) -> id^"."^(name_of_var more)

let listMethod env (var:reference) (ee:value list) =
  match var with
    | REF(id) ->
	(match id with
	   | "length" -> INT(List.length ee)
	   | _ -> invalid_arg("unknown array property: "^(name_of_var var)))
    | _ -> invalid_arg("unexpected variable "^(name_of_var var))

(* lookup a variable in the environment *)
let rec lookup (env:env) (var:reference) : value =
  match var with
    | REF(id) ->
	(try List.assoc id env
	 with Not_found -> invalid_arg("undefined variable "^id))
    | DOTREF(id,more) ->
	(try match (List.assoc id env) with
	   | ASSOC(env) -> lookup env more
	   | LIST(ee) -> listMethod env more ee
	   | _ -> invalid_arg(
	       "illegal dot reference. Variable "^id^" has no properties")
	 with Not_found -> invalid_arg("undefined variable "^id))

(* add/reset a variable in the environment *)
let rec extend (env:env) (var:reference) (v:value) : env =
  match var with
    | REF(id) -> (id,v)::env
    | DOTREF(id,more) ->
	if List.mem_assoc id env
	then match List.assoc id env with
	  | ASSOC(e) -> (id,ASSOC(extend e more v))::env
	  | _ -> invalid_arg("variable "^id^" previously bound to scalar."
			    ^" attempted to set properties")
	else (id,ASSOC(extend empty more v))::env

(* augment/overwrite references in e1 with those from e2 *)
let augment (e1:env) (e2:env) : env = e2 @ e1

let remove (env:env) (var:reference) : env =
  match var with
    | REF(id) -> List.remove_assoc id env
    | DOTREF(id,more) -> List.remove_assoc id env

(* get a printable version of a reference from the environment *)
let string_of_var (env:env) (var:reference) : string =
  match lookup env var with
    | LIST(vv) -> (name_of_var var)^":list"
    | STRING(s) -> s
    | INT(i) -> string_of_int i
    | BOOL(b) -> string_of_bool b
    | ASSOC(env) -> (name_of_var var)^":assoc"

let string_of_env (env:env) : string =
  let spaces n = String.make n ' ' in
  let rec soe ((env:env),(indent:int)) : string =
    let f s (a,b) =
      s^"\n"^(spaces indent)^a^
      (match b with
	 | LIST(vv) -> " : list"
	 | STRING(s) -> " : string = "^s
	 | INT(i) -> " : int = "^(string_of_int i)
	 | BOOL(b) -> " : bool = "^(string_of_bool b)
	 | ASSOC(e) -> " {"^(soe (e,(indent+2)))^"\n"^(spaces (indent+2))^"}")
    in List.fold_left f "" env
  in soe (env,0)

let string_of_range_expr (expr:range_expr) : string =
  match expr with
    | REF_RANGE(var) -> name_of_var var
    | INT_RANGE(i) -> string_of_int i

let rec string_of_pexpr (expr:prim_expr) : string =
  match expr with
    | REF_EXPR(var) -> name_of_var var
    | STR_EXPR(str) -> str
    | INT_EXPR(int) -> string_of_int int
    | BOOL_EXPR(b) -> string_of_bool b
    | LIST_EXPR(ee) ->
	let f s ex = s^", "^(string_of_pexpr ex)
	in "["^(List.fold_left f "" ee)^"]"
    | RANGE_EXPR(i,j) -> "["^(string_of_range_expr i)
	^".."^(string_of_range_expr j)^"]"

let rec string_of_expr (expr:expr) : string =
  let binop(op,e1,e2) =
    "("^(string_of_expr e1)^" "^op^" "^(string_of_expr e2)^")" in
  let unop(op,e1) = "("^op^" "^(string_of_expr e1)^")"
  in match expr with
    | PRIM(pexpr) -> string_of_pexpr pexpr
    | EQ(e1,e2) -> binop("==",e1,e2)
    | NE(e1,e2) -> binop("!=",e1,e2)
    | GT(e1,e2) -> binop(">",e1,e2)
    | GE(e1,e2) -> binop(">=",e1,e2)
    | LT(e1,e2) -> binop("<",e1,e2)
    | LE(e1,e2) -> binop(">=",e1,e2)
    | AND(e1,e2) -> binop("&&",e1,e2)
    | OR(e1,e2) -> binop("||",e1,e2)
    | ADD(e1,e2) -> binop("+",e1,e2)
    | SUB(e1,e2) -> binop("-",e1,e2)
    | MUL(e1,e2) -> binop("*",e1,e2)
    | DIV(e1,e2) -> binop("/",e1,e2)
    | MOD(e1,e2) -> binop("%",e1,e2)
    | NOT(e1) -> unop("!",e1)


let rec print_stmt stmt : unit =
  match stmt with
    | SCHMOO(t) -> print t
    | BLOCK(ss) -> List.iter print_stmt ss
    | SET_STMT(var,expr) ->
	print ("#set ("^(name_of_var var)^" = "^(string_of_expr expr)^")")
    | IF_STMT(expr,then_clause,else_clause_opt) ->
	print ("#if ("^(string_of_expr expr)^")");
	print_stmt then_clause;
	(match else_clause_opt with
	     None -> ()
	   | Some(else_clause) ->
	       print "#else"; print_stmt else_clause);
	print "#end"
    | FOR_STMT(index,pexpr,stmt) ->
	print ("#foreach ("^index^" in "^(string_of_pexpr pexpr)^")");
	print_stmt stmt;
	print "#end"
    | REF_STMT(var) -> print ("REF_STMT("^(name_of_var var)^")")
    | INC_STMT(pp) ->
	let s =
	  let f s ex = s^(string_of_pexpr ex)
	  in "#include ("^(List.fold_left f "" pp)^")"
	in print s
    | PARSE_STMT(pexpr) -> print ("#parse ("^(string_of_pexpr pexpr)^")")

let extendlist env (defns:(reference * value) list) : env =
  let f env (r,v) : env = extend env r v
  in List.fold_left f env defns
