open Vtl
open Buffer (* extensible string buffers *)

type estring = Buffer.t

let template_root = ref "/"

let parse_template (chan:in_channel) : template =
  let parse scan_f parse_f pos_f =
    let position (line,col) = ("near line "
                               ^ (string_of_int line)
                               ^ ", column "
                               ^ (string_of_int col)) in
    let lexbuf = Lexing.from_channel chan
    in
      try
	parse_f scan_f lexbuf
      with exn ->
	let pos = position (pos_f lexbuf) in
          begin match exn with
	    | Failure(msg) ->
		failwith ("program failure, while parsing " ^ pos ^ ": " ^ msg)
            | Parsing.Parse_error -> failwith("syntax error " ^ pos)
            | Sys_error msg -> failwith ("I/O error: " ^ msg)
            | _ -> raise exn
	  end
  in
    try
      Vtl_lexer.init();
      parse Vtl_lexer.scan Vtl_parser.template Vtl_lexer.position
    with Failure(msg) -> (* add some extra context for the parser guy! *)
      failwith (Vtl_states.contextAsString(Vtl_states.getContext())^":"^msg)

(* convert a range_expr to an integer. Raise Invalid_argument if
   range_expr does not evaluate to an integer.
*)
let int_of_range_expr (env:env) (expr:range_expr) : int =
  match expr with
    | INT_RANGE(i) -> i
    | REF_RANGE(var) -> match lookup env var with
	| INT(i) -> i
	| _ -> invalid_arg((name_of_var var)^" not a valid integer range expr")

(* get the value of one VTL expression *)
let rec value_of_pexpr (env:env) (expr:prim_expr) : value =
  match expr with
    | REF_EXPR(var) -> lookup env var
    | STR_EXPR(str) -> STRING(str)
    | INT_EXPR(int) -> INT(int)
    | BOOL_EXPR(b) -> BOOL(b)
    | LIST_EXPR(ee) -> LIST(List.map (value_of_pexpr env) ee)
    | RANGE_EXPR(ri,rj) ->
	let i = int_of_range_expr env ri in
	let j = int_of_range_expr env rj in
	let dx = if i < j then 1 else -1 in
	let stop = j+dx in
	let rec loop x = if x=stop then [] else (INT x)::(loop (x+dx))
	in LIST(loop i)

let rec eval_expr (env:env) (expr:expr) : value =
  let boolop(op,e1,e2) = op (bool_of_expr env e1) (bool_of_expr env e2) in
  let binop(op,e1,e2) = op (eval_expr env e1) (eval_expr env e2) in
  let mathop(op,e1,e2) =
    try op (int_of_expr env e1) (int_of_expr env e2)
    with Division_by_zero -> 0
  in match expr with
    | PRIM(pexpr) -> value_of_pexpr env pexpr
    | EQ(e1,e2) -> BOOL(binop((=),e1,e2))
    | NE(e1,e2) -> BOOL(binop((<>),e1,e2))
    | GT(e1,e2) -> BOOL(binop((>),e1,e2))
    | GE(e1,e2) -> BOOL(binop((>=),e1,e2))
    | LT(e1,e2) -> BOOL(binop((<),e1,e2))
    | LE(e1,e2) -> BOOL(binop((>=),e1,e2))
    | AND(e1,e2) -> BOOL(boolop((&&),e1,e2))
    | OR(e1,e2) -> BOOL(boolop((||),e1,e2))
    | ADD(e1,e2) -> INT(mathop((+),e1,e2))
    | SUB(e1,e2) -> INT(mathop((-),e1,e2))
    | MUL(e1,e2) -> INT(mathop(( * ),e1,e2))
    | DIV(e1,e2) -> INT(mathop((/),e1,e2))
    | MOD(e1,e2) -> INT(mathop((mod),e1,e2))
    | NOT(e) -> BOOL(not(bool_of_expr env e))

(* evaluate and convert an expr to boolean according to VTL rules *)
and bool_of_expr (env:env) (expr:expr) : bool =
  match (eval_expr env expr) with
    | BOOL(b) -> b
    | INT(i) -> i != 0
    | _ -> true

(* evaluate and convert an expr to boolean according to VTL rules *)
and int_of_expr (env:env) (expr:expr) : int =
  match (eval_expr env expr) with
    | BOOL(b) -> if b then 1 else 0
    | INT(i) -> i
    | STRING(s) -> (try int_of_string s with Invalid_argument(m) -> 0)
    | _ -> 0

(* pull in the included text files and append them to the supplied string *)
let eval_include ((env:env),(str:estring)) (expr:prim_expr) : env * estring =
  (* read all input from a file into one string *)
  let rec readfile chan : unit =
    try
      let line = (input_line chan)^"\n"
      in
	Buffer.add_string str line;
	readfile chan
    with End_of_file -> () in
  let s = match value_of_pexpr env expr with
    | STRING(s) ->
	let fname = !template_root^"/"^s
	in
	  (try
	     let chan = (open_in fname)
	     in readfile chan
	   with Sys_error(msg) ->
	     Buffer.add_string str ("<!-- #include: "^fname^" not found. -->"))
    | _ -> Buffer.add_string str ("<!-- #include: "^(Vtl.string_of_pexpr expr)
				  ^" does not evaluate to string -->")
  in env,str

(* evaluate one VTL statement *)
let rec eval_stmt ((env:env),(str:estring)) (stmt:stmt) : env * estring =
  (* eval a foreach statement, convert singletons in the pexpr to a list.
     The index variable should only have local scope. The VTL spec says
     that $velocityCount should hold the current array index.
  *)
  let eval_for_stmt (index,pexpr,stmt) env : env * estring =
    try
      let var = REF(index) in
      let counterVar = REF("$velocityCount") in
      let f (env,s,counter) item =
	let e = extend (extend env var item) counterVar (INT counter) in
	let e,s = eval_stmt (e,s) stmt
	in (remove (remove e var) counterVar),s,(counter+1) in
      let items = match pexpr with
	| REF_EXPR(var) ->
	    (match (lookup env var) with
	       | LIST(vs) -> vs
	       | v -> [v])
	| LIST_EXPR(pp) -> List.map (value_of_pexpr env) pp
	| STR_EXPR(s) -> [STRING(s)]
	| INT_EXPR(i) -> [INT(i)]
	| BOOL_EXPR(b) -> [BOOL(b)]
	| RANGE_EXPR(i,j) ->
	    match value_of_pexpr env (RANGE_EXPR(i,j)) with
	      | LIST(vv) -> vv
	      | _ -> invalid_arg("internal error on range expr.") in
      let env,str,_ = List.fold_left f (env,str,0) items
      in env,str
    with Invalid_argument(msg) ->
      (* append error message and return new env and buffer *)
      Buffer.add_string str ("<!-- "^msg^" in FOREACH -->");
      env,str
  in match stmt with
    | SCHMOO(t) -> Buffer.add_string str t; env,str
    | BLOCK(ss) -> List.fold_left eval_stmt (env,str) ss
    | SET_STMT(var,expr) -> (extend env var (eval_expr env expr)),str
    | IF_STMT(expr,then_clause,else_clause_opt) ->
	(try
	   if (try bool_of_expr env expr with _ -> false)
	   then eval_stmt (env,str) then_clause
	   else (match else_clause_opt with
		   | None -> env,str
		   | Some(else_clause) -> eval_stmt (env,str) else_clause)
	 with Invalid_argument(msg) ->
	   Buffer.add_string str ("<!-- "^msg^" in IF -->");
	   env,str)
    | FOR_STMT(index,pexpr,stmt) -> eval_for_stmt (index,pexpr,stmt) env
    | REF_STMT(var) ->
	let s =
	  try string_of_var env var
	  with Invalid_argument(msg) -> "<!-- "^msg^" -->"
	in
	  Buffer.add_string str s;
	  env,str
    | INC_STMT(pp) -> List.fold_left eval_include (env,str) pp
    | PARSE_STMT(p) -> eval_parse env str p
and eval_parse (env:env) (str:estring) (expr:prim_expr) : env * estring =
  (* parse the file referenced by [expr] to arrive at a new list of stmts.
     fold eval_stmt back over this to include any modifications to the
     environment *)
  try
    let fname = match value_of_pexpr env expr with
      | STRING(s) -> !template_root^"/"^s
      | _ -> invalid_arg("#parse: expected string argument") in
    let stmts =
      try parse_template (open_in fname)
      with Sys_error(_) -> invalid_arg("#parse: can't open "^fname)
    in List.fold_left eval_stmt (env,str) stmts
  with Invalid_argument(msg) ->
    Buffer.add_string str ("<!-- "^msg^" -->"); env,str

(*
 * evaluate a whole template with context of supplied environment. The
 * environment should have been filled by the web-application. Stmts
 * can alter the env. The returned string has all substitutions and
 * is free of VTL-specific code.
*)
let eval_template (env:env) (stmts:template) : string =
  let _,document =
    List.fold_left eval_stmt (env,(Buffer.create (1024*32))) stmts
  in Buffer.contents document

(*
 * run - parse a [template] file from the [root] template directory
 * and make substitutions from the supplied environment.
 *)
let run (env:env) (root:string) (file:string) : string =
  template_root := root;
  let name = (!template_root^"/"^file) in
  let template =
    try parse_template (open_in name)
    with Sys_error(msg) -> []
  in
    if template=[] then "VTL Template not found: "^name
    else
      try eval_template env template
      with Invalid_argument(msg) -> (msg^"\n"
				     ^"The environment contains:\n"
				     ^(string_of_env env))
