{
  open Vtl_parser  (* token definitions *)
  open Vtl_states  (* setContext() *)
  
  (* little helpers *)

  let pos   = Lexing.lexeme_start
  let last  = Lexing.lexeme_end
  let get   = Lexing.lexeme

  let error(msg) = failwith(msg)

  (* simple minded definitions for line counting. Whenever we
   * we scan a newline we call newline().
   *)

  let line         = ref 1
  let startOfLine  = ref 0

  (* table of unique strings *)
  
  let init () =
    line          := 1;
    startOfLine   := 0;
    setContext(SchmooContext)

  let newline lexbuf =
    line := !line + 1;
    startOfLine := Lexing.lexeme_start lexbuf

  (* actual (line, column) position *)
  let position lexbuf = (!line,Lexing.lexeme_start lexbuf - !startOfLine)

  let suffix str n = String.sub str n (String.length str - n)
  let prefix str n = String.sub str 0 (String.length str - n)

  let trim_var v =
    if String.contains v '{' then "$"^(suffix v 2) else v
  
}

(* regexp declarations *)

let     ws     = [' ' '\t' '\r']
let     nl     = '\n'
let     ident  = ['a'-'z' 'A'-'Z'] + ['a'-'z' 'A'-'Z' '0'-'9' '-' '_'] *
let     dot    = '.'
let     var    = '$' '{'? ident
let     integer= ['0'-'9'] +
let     comment= "##" [^ '\n'] *
let     xset    = '#' ("set"|"SET")
let     xinclude= '#' ("include"|"INCLUDE")
let     xparse  = '#' ("parse"|"PARSE")
let     xforeach= '#' ("foreach"|"FOREACH")
let     xelse  = '#' ("else"|"ELSE")
let     xend   = '#' ("end"|"END")
let     xif    = '#' ("if"|"IF")

(* ------------------------------------------------------------------
   different lexers - the active lexer is controlled by
   the parser!
   ------------------------------------------------------------------ *)

(* inside a directive, i.e. after the # *)
rule in_directive = parse
      eof			{ EOF }
   
   (* whitespace and newlines *)
   | nl	  		        { newline lexbuf; in_directive lexbuf }
   | ws	+                       { in_directive lexbuf }

   | "in"			{ IN       }
   | "IN"			{ IN       }
   | "TRUE"			{ TRUE     }
   | "FALSE"			{ FALSE    }
   | '='			{ ASSIGN   }
   | '"'			{ string1 lexbuf }
   | '\''			{ string2 lexbuf }
   | '('			{ LP }
   | ')'			{ RP }
   | '['			{ LB }
   | ']'			{ RB }
   | ','			{ COMMA }
   | ".."			{ DOTDOT }
   | '-'			{ MINUS }
   | '+'			{ PLUS }
   | '*'			{ MULT }
   | '/'			{ DIV }
   | '%'			{ MOD }
   | '<'			{ LT }
   | '>'			{ GT }
   | "<="			{ LE }
   | ">="			{ GE }
   | "=="			{ EQ }
   | "!="			{ NE }
   | "!"			{ NOT }
   | "&&"			{ AND }
   | "||"			{ OR }
   | integer                    { Vtl_parser.INT(int_of_string (get lexbuf)) }

   (* variable references *)
   | var                        { ID(trim_var(get lexbuf)) }
   | ident                      { ID(get lexbuf) }
   | dot                        { DOT }
   | '}'                        { in_directive lexbuf }

   (* everything else is illegal *)
   | _				{ error "illegal character" }

(* outside of directives, i.e. we scan raw text *)
and in_schmoo = parse
   | eof		 { EOF                                   }
   | nl			 { newline lexbuf; TEXT("\n")            }
   | '\\' '$'            { TEXT(get lexbuf) (*escaped reference*)}
   | var                 { setContext(ReferenceContext);
			   ID(trim_var(get lexbuf))              }
   | comment '\n'        { newline lexbuf; in_schmoo lexbuf      }
   | [^ '#' '$' '\n' ] * { TEXT(get lexbuf)                      }
   | xset                { setContext(DirectiveContext); SET     }
   | xinclude            { setContext(DirectiveContext); INCLUDE }
   | xparse              { setContext(DirectiveContext); PARSE   }
   | xif                 { setContext(DirectiveContext); IF      }
   | xforeach            { setContext(DirectiveContext); FOREACH }
   | xelse               { ELSE                                  }
   | xend                { END                                   }
   | '#'                 { TEXT(get lexbuf)                      }
   | _			 { error ("illegal character `"
				  ^(get lexbuf)^"' in schmoo")   }

(* inside a variable reference, i.e. after the $ *)
and in_reference = parse
   | nl			       { newline lexbuf; in_schmoo lexbuf }
   | '}'                       { in_schmoo lexbuf }
   | dot                       { DOT }
   | ident                     { ID(get lexbuf) }
   | ws +                      { TEXT(get lexbuf) }
   | _			       { error ("illegal character `"
					^(get lexbuf)
					^"' in variable reference")
			       }

and string1 = parse
   | [^ '"' '\n']* '"'          { Vtl_parser.STRING(prefix (get lexbuf) 1)}
   | [^ '"' '\n']* eof		{ error "unterminated string"   }
   | [^ '"']* nl	        { error "newline in string is unsupported"}
   | _				{ error "illegal char in string"}

and string2 = parse
   | [^ '\'' '\n']* '\''        { Vtl_parser.STRING(prefix (get lexbuf) 1)}
   | [^ '\'' '\n']* eof		{ error "unterminated string"   }
   | [^ '\'']* nl		{ error "newline in string is unsupported"}
   | _				{ error "illegal char in string"}

(* From the outside the scanner is always invoced through [scan]
 * which will respect the currently active context 
 *)

{
let scan buffer =
    match getContext() with
    | SchmooContext	    -> in_schmoo     buffer
    | DirectiveContext      -> in_directive  buffer
    | ReferenceContext      -> in_reference  buffer
}
