/*
 * Velocity Template Language Parser for Vtl module.
 */
%{
open Vtl
open Vtl_states
%}

%token <int> INT
%token <string> STRING
%token <string> ID
%token <string> TEXT
%token LP RP COMMA LB RB
%token DOT IF ELSE FOREACH IN SET ASSIGN END DOTDOT
%token INCLUDE PARSE
%token TRUE FALSE
%token EQ NE GT GE LT LE
%token AND OR NOT
%token MINUS UMINUS
%token EOF
%token MULT DIV MOD PLUS
%left OR AND EQ NE GT GE LT LE
%left MULT DIV MOD PLUS
%left NOT
%left MINUS
%nonassoc UMINUS

%start template
%type <Vtl.template> template
%type <int> integer
%%

template: stmts EOF   { List.rev $1 }
;

stmts: /**/           { [] }
     | stmts stmt     { $2::$1 }

var : ID               { Vtl.REF($1) }
    | ID DOT var       { Vtl.DOTREF($1,$3) }
;

stmt: if_expr then_clause else_clause    { setContext(SchmooContext);
					   IF_STMT($1,$2,$3) }
    | for_expr stmts END                 { setContext(SchmooContext);
					   FOR_STMT(fst($1),snd($1),
						    BLOCK(List.rev $2)) }
    | SET LP var ASSIGN expr RP          { setContext(SchmooContext);
					   SET_STMT($3,$5)    }
    | TEXT                               { Vtl.SCHMOO($1)         }
    | var                                { setContext(SchmooContext);
					   REF_STMT($1)       }
    | INCLUDE LP path_list RP            { setContext(SchmooContext);
					   INC_STMT($3)       }
    | PARSE LP path_item RP              { setContext(SchmooContext);
					   PARSE_STMT($3)       }
/*
 * #if(expr) stmts [#else stmts] #end
 *
 * Broken apart so that we can set the context for statement lexing
 */
if_expr: IF LP expr RP   { setContext(SchmooContext); $3 }
then_clause: stmts       { BLOCK(List.rev $1) }
else_clause: END         { None }
    | ELSE stmts END     { Some(BLOCK(List.rev $2)) }

/*
 * #foreach(id in prim_expr)
 *
 * range should evaluate to a list.
 * Broken apart so that we can set the context for statement lexing
 */
for_expr: FOREACH LP ID IN prim_expr RP  { setContext(SchmooContext); ($3,$5) }

integer: INT                          { $1   }
  | MINUS INT %prec UMINUS            { - $2 }
;

expr: prim_expr            { Vtl.PRIM($1)  }
    | LP expr RP           { $2 }
    | expr EQ expr         { Vtl.EQ($1,$3) }
    | expr NE expr         { Vtl.NE($1,$3) }
    | expr GT expr         { Vtl.GT($1,$3) }
    | expr GE expr         { Vtl.GE($1,$3) }
    | expr LT expr         { Vtl.LT($1,$3) }
    | expr LE expr         { Vtl.LE($1,$3) }
    | expr AND expr        { Vtl.AND($1,$3) }
    | expr OR expr         { Vtl.OR($1,$3) }
    | expr PLUS expr       { Vtl.ADD($1,$3) }
    | expr MINUS expr      { Vtl.SUB($1,$3) }
    | expr MULT expr       { Vtl.MUL($1,$3) }
    | expr DIV expr        { Vtl.DIV($1,$3) }
    | expr MOD expr        { Vtl.MOD($1,$3) }
    | NOT expr             { Vtl.NOT($2) }
;

/* support [x..y] where x and y are integer variables or integers */

range_expr: var             { Vtl.REF_RANGE($1) }
          | integer         { Vtl.INT_RANGE($1) }

prim_expr: integer          { Vtl.INT_EXPR($1) }
         | STRING           { Vtl.STR_EXPR($1) }
         | TRUE             { Vtl.BOOL_EXPR(true) }
         | FALSE            { Vtl.BOOL_EXPR(false) }
         | var              { Vtl.REF_EXPR($1) }
         | LB list RB       { Vtl.LIST_EXPR($2) }
         | LB range_expr DOTDOT range_expr RB
                            { Vtl.RANGE_EXPR($2,$4) }
;

list: /**/                             { [] }
    | non_empty_list                   { $1 }

non_empty_list: prim_expr              { [$1] }
              | prim_expr COMMA list   { $1::$3 }

/* #include( path-list )
*/
path_list: path_item              { [$1] }
	 | path_item COMMA list   { $1::$3 }

path_item: STRING           { Vtl.STR_EXPR($1) }
	 | var              { Vtl.REF_EXPR($1) }

