(*
 * app_server - combine wserver with the VTL run function.
 *
 * app_server port template_root services
 *)

type uri = string
type name = string
type value = string
type attr = name * value
type params = attr list

type srvc_func = (string * params) -> Vtl.env

(* template name * service function *)
type service = string * srvc_func

(* services association defines the whole application *)
type services = (name * service) list

type delayedServices = (name * (unit->service)) list

let hexa_digit x =
  if x >= 10 then Char.chr (Char.code 'A' + x - 10)
  else Char.chr (Char.code '0' + x)
;;

let hexa_val conf =
  match conf with
    '0'..'9' -> Char.code conf - Char.code '0'
  | 'a'..'f' -> Char.code conf - Char.code 'a' + 10
  | 'A'..'F' -> Char.code conf - Char.code 'A' + 10
  | _ -> 0
;;

let decode s =
  let rec need_decode i =
    if i < String.length s then
      match s.[i] with
        '%' | '+' -> true
      | _ -> need_decode (succ i)
    else false
  in
  let rec compute_len i i1 =
    if i < String.length s then
      let i =
        match s.[i] with
          '%' when i + 2 < String.length s -> i + 3
        | _ -> succ i
      in
      compute_len i (succ i1)
    else i1
  in
  let rec copy_decode_in s1 i i1 =
    if i < String.length s then
      let i =
        match s.[i] with
          '%' when i + 2 < String.length s ->
            let v = hexa_val s.[i + 1] * 16 + hexa_val s.[i + 2] in
            s1.[i1] <- Char.chr v; i + 3
        | '+' -> s1.[i1] <- ' '; succ i
        | x -> s1.[i1] <- x; succ i
      in
      copy_decode_in s1 i (succ i1)
    else s1
  in
  let rec strip_heading_and_trailing_spaces s =
    if String.length s > 0 then
      if s.[0] == ' ' then
        strip_heading_and_trailing_spaces
          (String.sub s 1 (String.length s - 1))
      else if s.[String.length s - 1] == ' ' then
        strip_heading_and_trailing_spaces
          (String.sub s 0 (String.length s - 1))
      else s
    else s
  in
  if need_decode 0 then
    let len = compute_len 0 0 in
    let s1 = String.create len in
    strip_heading_and_trailing_spaces (copy_decode_in s1 0 0)
  else s
(***************** end of code taken from wserver.ml ******************)

(* split a string on a seperater [sep] character and return a
   list of the pieces sans char *)
let rec split (sep:char) (s:string) : string list =
  try
    let index = String.index s sep in
    let hd = String.sub s 0 index in
    let rest = String.sub s (index+1) ((String.length s)-index-1)
    in hd::(split sep rest)
  with _ -> if s="" then [] else [s]

let splitQueryArgs (query:string) : params =
  let params = split '&' query in
  let mkattr s =
    let ss = split '=' s
    in
      (try (List.hd ss) with _ -> ""),
      (try (List.hd (List.tl ss)) with _ -> "")
  in List.map mkattr params

let rec replace (c:char) (s:string) str =
  try
    let index = String.index str c in
      if index = 0 then str
      else (
	let len = String.length str in
	let hd = String.sub str 0 index in
	let tl = String.sub str (index+1) ((len-index)-1) in
	  hd^s^(replace c s tl)
      )
  with _ -> str

  (* split a request (URL with query parameters) into the path and params *)
let splitRequest (request:uri) : string * params =
  let ss = split '?' (Wserver.decode request) in
  let path,query =
    (try (List.hd ss) with _ -> ""),
    (try (List.hd (List.tl ss)) with _ -> "") in
  let env = splitQueryArgs query
  in path,env

(* build a query URI from a list of attributes *)
let queryUri(base,attrs) =
  let rec render = function
    | [] -> ""
    | (n,v)::[] -> n^"="^v
    | (n,v)::rest -> n^"="^v^"&"^(render rest)
  in base^"?"^(render attrs)

(* return a value from the params *)
let getEnv (env:params) name : string =
  try List.assoc name env with Not_found -> ""

(* return all values from the params as a list *)
let getEnvAll (env:params) name : string list =
  let all = List.filter (fun (x,y) -> x=name) env
  in List.map (fun (x,y) -> y) all

let handleStaticRequest chan template_root (request:string) =
  let name = (template_root^"/"^request)
  in
    try
      let ichan = open_in name
      in
	try
	  while true do
	    output_byte chan (input_byte ichan)
	  done
	with End_of_file -> ()
    with Sys_error(msg) -> raise Not_found

(* Server that allows custom error handling *)
type error_handler = string*params -> exn -> unit
let customServer (port:int) (template_root:string) (globalEnv:Vtl.env)
  (services:services) (not_found_srvc:service) (error_handler:error_handler) =
  let serve chan inetaddr request =
(*    Printf.eprintf "GET %s\n" request; *)
    let path,params = splitRequest request in
    let doRequest(service) =
      let template_name,svc_func = service in
      let env = Vtl.augment globalEnv (svc_func(path,params)) in
	(* things in globalEnv are overwitten *)
      let doc = Vtl_engine.run env template_root template_name
      in
	output_string chan doc
    in
      try
	let service = List.assoc path services
	in
	  try doRequest(service)
	  with exn -> error_handler (path,params) exn
      with Not_found ->
	(* can't find a VTL binding, maybe it's a static request...*)
	try handleStaticRequest chan template_root request
	with Not_found -> doRequest(not_found_srvc)
  in
    try Wserver.f_st port (serve stdout)
    with Sys_error msg -> Printf.eprintf "server: fatal error: %s\n" msg

(***************************************************************************
 * server
 *
 * port - the HTTP port on which this app will run
 * template_root - root directory for finding VTL templates for this server
 * services - the site map. binds path names to service functions
 * not_found_srvc - called when a supplied URI has no binding in services.
 *          It receives the path name and parameters of the failed request.
 ***************************************************************************
*)
let server (port:int) (template_root:string) (services:services)
  (not_found_srvc:service) =
  let error_handler (path,params) exn =
    Printf.printf "Application error in %s\n" path;
    Printf.printf "Exception: %s\n" (Printexc.to_string exn);
    flush stdout;
    ()
  in customServer port template_root Vtl.empty services
       not_found_srvc error_handler
