(****************************************************************)
(* here starts a function taken from Wserver, a module from the
 * geneology project by Daniel de Rauglaudre found here
 * ftp://ftp.inria.fr/INRIA/Projects/cristal/Daniel.de_Rauglaudre/Tools/
 *
 * This version has been modified from the original so that it does not
 * fork as many child processes. We had trouble hunting them all down
 * when we wanted to kill the application.
 *)


open Unix;;

let hexa_digit x =
  if x >= 10 then Char.chr (Char.code 'A' + x - 10)
  else Char.chr (Char.code '0' + x)
;;

let hexa_val conf =
  match conf with
    '0'..'9' -> Char.code conf - Char.code '0'
  | 'a'..'f' -> Char.code conf - Char.code 'a' + 10
  | 'A'..'F' -> Char.code conf - Char.code 'A' + 10
  | _ -> 0
;;

let decode s =
  let rec need_decode i =
    if i < String.length s then
      match s.[i] with
        '%' | '+' -> true
      | _ -> need_decode (succ i)
    else false
  in
  let rec compute_len i i1 =
    if i < String.length s then
      let i =
        match s.[i] with
          '%' when i + 2 < String.length s -> i + 3
        | _ -> succ i
      in
      compute_len i (succ i1)
    else i1
  in
  let rec copy_decode_in s1 i i1 =
    if i < String.length s then
      let i =
        match s.[i] with
          '%' when i + 2 < String.length s ->
            let v = hexa_val s.[i + 1] * 16 + hexa_val s.[i + 2] in
            s1.[i1] <- Char.chr v; i + 3
        | '+' -> s1.[i1] <- ' '; succ i
        | x -> s1.[i1] <- x; succ i
      in
      copy_decode_in s1 i (succ i1)
    else s1
  in
  let rec strip_heading_and_trailing_spaces s =
    if String.length s > 0 then
      if s.[0] == ' ' then
        strip_heading_and_trailing_spaces
          (String.sub s 1 (String.length s - 1))
      else if s.[String.length s - 1] == ' ' then
        strip_heading_and_trailing_spaces
          (String.sub s 0 (String.length s - 1))
      else s
    else s
  in
  if need_decode 0 then
    let len = compute_len 0 0 in
    let s1 = String.create len in
    strip_heading_and_trailing_spaces (copy_decode_in s1 0 0)
  else s
;;

let special x = List.mem x ['='; '&'; '"'];;

let encode s =
  let rec need_code i =
    if i < String.length s then
      match s.[i] with
        ' ' -> true
      | x -> if special x then true else need_code (succ i)
    else false
  in
  let rec compute_len i i1 =
    if i < String.length s then
      let i1 = if special s.[i] then i1 + 3 else succ i1 in
      compute_len (succ i) i1
    else i1
  in
  let rec copy_code_in s1 i i1 =
    if i < String.length s then
      let i1 =
        match s.[i] with
          ' ' -> s1.[i1] <- '+'; succ i1
        | c ->
            if special c then
              begin
                s1.[i1] <- '%';
                s1.[i1 + 1] <- hexa_digit (Char.code c / 16);
                s1.[i1 + 2] <- hexa_digit (Char.code c mod 16);
                i1 + 3
              end
            else begin s1.[i1] <- c; succ i1 end
      in
      copy_code_in s1 (succ i) i1
    else s1
  in
  if need_code 0 then
    let len = compute_len 0 0 in copy_code_in (String.create len) 0 0
  else s
;;

let print_exc exc =
  match exc with
    Unix_error (err, fun_name, arg) ->
      prerr_string "\"";
      prerr_string fun_name;
      prerr_string "\" failed";
      if String.length arg > 0 then
        begin
          prerr_string " on \""; prerr_string arg; prerr_string "\""; ()
        end;
      prerr_string ": ";
      prerr_endline (error_message err);
      flush Pervasives.stderr;
      ()
  | _ ->
      try prerr_endline (Printexc.print raise exc) with
        _ -> ()
;;

let rec request s =
  parser
  [< ''\n' | ' ' >] -> s | [< 'c; a = request (s ^ String.make 1 c) >] -> a
;;

let get = parser [< ''G'; ''E'; ''T'; '' '; ''/'; a = request "" >] -> a;;

let timeout tmout spid _ =
  Unix.kill spid Sys.sigkill;
  Printf.printf "<head><title>Time out</title></head>\n";
  Printf.printf "<body><h1>Time out</h1>\n";
  Printf.printf "Computation time > %d second(s)\n" tmout;
  Printf.printf "</body>\n";
  flush Pervasives.stdout;
  exit 2
;;

let rec accept_connection tmout f s =
  let (t, addr) = accept s in
  match
    try Some (fork ()) with
      _ -> None
  with
    Some 0 ->
      if fork () <> 0 then exit 0;
      close s;
      dup2 t stdout;
      dup2 t stdin;
      Printf.printf "HTTP/1.0 200 OK\n";
      Printf.printf "Content-type: text/html\n\n";
      let spid = fork () in
      if spid <> 0 then
        begin
          if tmout > 0 then
            (let _ = Sys.signal Sys.sigalrm (Sys.Signal_handle (timeout tmout spid)) in
             let _ = alarm tmout in
             let _ = Unix.wait () in
             let _ = Sys.signal Sys.sigalrm Sys.Signal_default
	     in ());
          (try close t with _ -> ());
          exit 0
        end
      else
        begin
          begin try
            let request = get (Stream.of_channel Pervasives.stdin) in
            f addr request;
            flush Pervasives.stdout;
            shutdown t SHUTDOWN_ALL;
            ()
          with
            e ->
              try Printexc.print raise e with
                _ -> ()
          end;
          begin try flush Pervasives.stdout with
            _ -> ()
          end;
          begin try flush Pervasives.stderr with
            _ -> ()
          end;
          begin try close t with
            _ -> ()
          end;
          exit 0;
          ()
        end
  | Some id -> close t; 
      begin let _ = waitpid [] id in () end; ()
  | None ->
      close t; Printf.eprintf "Fork failed\n"; flush Pervasives.stderr; ()
;;

  
let f port tmout g =
  try
    let s = socket PF_INET SOCK_STREAM 0 in
    setsockopt s SO_REUSEADDR true;
    bind s (ADDR_INET (inet_addr_any, port));
    listen s 1;
    let tm = localtime (time ()) in
    Printf.eprintf "Ready %02d/%02d/%02d %02d:%02d...\n" tm.tm_mday
      (succ tm.tm_mon) tm.tm_year tm.tm_hour tm.tm_min;
    flush Pervasives.stderr;
    while true do
      try accept_connection tmout g s with
        exc -> print_exc exc
    done;
    ()
  with
    exc -> print_exc exc
;;

let rec accept_connection_st f s =
  let (t, addr) = accept s in
    dup2 t stdout;
    dup2 t stdin;
    Printf.printf "HTTP/1.0 200 OK\n";
    Printf.printf "Content-type: text/html\n\n";
    let request = get (Stream.of_channel Pervasives.stdin) in
      (try 
	 f addr request;
       with exn -> 
	 close t;
	 close s;
	 raise exn);
      flush Pervasives.stdout;
      shutdown t SHUTDOWN_ALL;
      (try flush Pervasives.stdout
       with _ -> ());
      (try flush Pervasives.stderr
       with _ -> ());
      (try close t
       with _ -> ());
      ()

let f_st port g =
  let s = socket PF_INET SOCK_STREAM 0 in
    setsockopt s SO_REUSEADDR true;
    bind s (ADDR_INET (inet_addr_any, port));
    listen s 1;
    let tm = localtime (time ()) in
      Printf.eprintf "Ready %02d/%02d/%02d %02d:%02d...\n" tm.tm_mday
	(succ tm.tm_mon) tm.tm_year tm.tm_hour tm.tm_min;
      flush Pervasives.stderr;
      while true do
	accept_connection_st g s 
      done;
      flush Pervasives.stderr;
      ()
;;
